/******************************************************************************
Utility to convert a LTP flow to a TCP flow for LTP analysis via tcptrace.
Utility Functions for Encapsulation

Copyright (C) 2013  Samuel Jero <sj323707@ohio.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Author: Samuel Jero <sj323707@ohio.edu>
Date: 06/2013

Notes:
	1)Only handles one LTP "connection". There isn't a good way to separate
		different LTP "connections" from new sessions of the same "connection".
		Use Tcpdump filters to separate connections. Libpcap filtering could also
		be added in ltp2tcp.
	2)Uses some special types from Linux (u_char, u_int32_t)
******************************************************************************/
#include "ltp2tcp.h"
#include "checksums.h"



/*Encapsulation Selection*/
void encap_sel(char* string){
	if(strcmp(string, "udp")==0 || strcmp(string,"UDP")==0){ /*UDP*/
		state.en_ops=&udp_encap;
		return;
	}
	if(strcmp(string, "dccp")==0 || strcmp(string,"DCCP")==0){ /*DCCP*/
		state.en_ops=&dccp_encap;
		return;
	}
	if(strcmp(string, "sll")==0 || strcmp(string,"SLL")==0){ /*SLL (Linux Cooked Capture)*/
			state.en_ops=&sll_encap;
			return;
		}
	printf("Encapsulation type: %s is not supported\n", string);
	exit(1);
return;
};

/*Fill the encapsulation structure*/
int fill_eip4_encap(struct eip4_en_p *eip, const u_char* data, int dlen, struct pcap_pkthdr *h){
	/*safety check*/
	if(eip==NULL || data==NULL || h==NULL || dlen < sizeof(struct ether_header)+sizeof(struct iphdr)){
		dbgprintf(1, "Error: Ethernet, IPv4 Encapsulation method given bad data!\n");
		return -1;
	}

	if(eip->first==0){
		/* First time, allocate memory and copy libpcap header and encap headers
		 * this guarantees the IP "direction" of the encap headers */
		memcpy(&eip->header, h, sizeof(struct pcap_pkthdr));
		memcpy(eip->od, data,sizeof(struct ether_header)+sizeof(struct iphdr));
		eip->first=1;
	}else{
		/* Just update the libpcap header (and associated timestamp)*/
		memcpy(&eip->header, h, sizeof(struct pcap_pkthdr));
	}
	return 0;
}

/* encapsulation manipulation after conversion */
int eip4_post(struct eip4_en_p *eip, int tlen, u_char* data){
	struct iphdr *iph;
	struct tcphdr *tcph;

	/* Move data pointer to start of IPv4 header*/
	data+=sizeof(struct ether_header);

	/*Determine if the given length is reasonable*/
	if((tlen+sizeof(struct iphdr)) > 0xFFFF){
			dbgprintf(1, "Error: Given TCP header+data length is too large for an IPv4 packet!\n");
			return -1;
	}

	/*Adjust IPv4 header to account for packet's total length*/
	iph=(struct iphdr*)data;
	iph->tot_len=htons(iph->ihl*4+tlen);

	/*Compute IPv4 Checksum*/
	iph->check=0;
	iph->check=ipv4_chksum(data,iph->ihl*4);

	/*Compute TCP Checksum*/
	data+=iph->ihl*4;
	tcph=(struct tcphdr*)data;
	tcph->check=0;
	tcph->check=ipv4_pseudohdr_chksum(data, tlen, (u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);
	return 0;
}

/* Create a TCP three-way handshake */
int eip4_handshake(struct eip4_en_p *eip, struct pcap_pkthdr *h){
	struct iphdr 		*iph;
	struct tcphdr 		*tcph;
	u_char 				*data;
	u_char 				*ptr;
	struct pcap_pkthdr	nh;
	u_int32_t			temp;

	/*Safety Check*/
	if(h==NULL || state.en_priv==NULL || eip==NULL){
		dbgprintf(1, "Error: Ethernet, IPv4 Encapsulation handshake method given bad data!\n");
		return -1;
	}

	/*create new libpcap header*/
	memcpy(&nh, h, sizeof(struct pcap_pkthdr));

	/*create buffer for new packet*/
	ptr=data=malloc(MAX_PACKET);
	if(data==NULL){
		dbgprintf(0,"Error: Couldn't allocate Memory\n");
		exit(1);
	}

	/* 1)Create Syn Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr)+4;
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr)+4;
		nh.ts.tv_usec-=3000; /*Time comes from the first packet received, so make these packets earlier*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(ptr, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Adjust IP header*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr)+4);

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=6;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=0;
		tcph->syn=1;
		tcph->rst=0;
		tcph->ack=0;

		/*Initialize Sequence and Acknowledgment Numbers and Window*/
		tcph->seq=htonl(state.seq_num++);
		tcph->ack_seq=htonl(0);
		tcph->window=htons(WIN_FACTOR);

		/* Add SACK permitted option*/
		ptr+=sizeof(struct tcphdr);
		*ptr=4;
		ptr++;
		*ptr=2;

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,tcph->doff*4,
										(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/*Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);


	/* 2)Create Syn,Ack Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr)+4;
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr)+4;
		nh.ts.tv_usec+=1000; /*This packet is 1/3rd closer to the first packet then the previous packet created*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(data, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Adjust IP header, including swapping source and destination*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		temp=iph->saddr;
		iph->saddr=iph->daddr;
		iph->daddr=temp;
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr)+4);

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=6;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=0;
		tcph->syn=1;
		tcph->rst=0;
		tcph->ack=1;

		/*Initialize Sequence and Acknowledgement Numbers and Window*/
		tcph->seq=htonl(state.ack_num++);
		tcph->ack_seq=htonl(state.seq_num);
		tcph->window=htons(WIN_FACTOR);

		/* Add SACK permitted option*/
		ptr+=sizeof(struct tcphdr);
		*ptr=4;
		ptr++;
		*ptr=2;

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,tcph->doff*4 ,
									(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/*Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);

	/* 3)Create Ack Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.ts.tv_usec+=1000; /*This packet is 2/3rds between SYN and first packet*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(data, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Adjust IP header*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr));

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=5;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*Initialize Sequence and Acknowledgement numbers and window*/
		tcph->seq=htonl(state.seq_num++);
		tcph->ack_seq=htonl(state.ack_num);
		tcph->window=htons(WIN_FACTOR);

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,tcph->doff*4,
								(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/*Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);
	return 0;
}

/* Create a TCP ending handshake */
int eip4_fin(struct eip4_en_p *eip){
	struct iphdr 		*iph;
	struct tcphdr 		*tcph;
	u_char 				*data;
	u_char 				*ptr;
	struct pcap_pkthdr	nh;
	u_int32_t			temp;

	/*Safety Check*/
	if(eip==NULL){
		dbgprintf(1,"Error: Ethernet, IPv4 Encapsulation Finish method given invalid data!\n");
		return -1;
	}

	/*copy the libpcap header from private data area*/
	memcpy(&nh, &eip->header, sizeof(struct pcap_pkthdr));

	/*create buffer for new packet*/
	ptr=data=malloc(MAX_PACKET);
	if(data==NULL){
		dbgprintf(0,"Error: Couldn't allocate Memory\n");
		exit(1);
	}

	/* 1)Create Fin Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.ts.tv_usec+=1000; /*Time is from the last packet in the capture; make this packet after that packet*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(ptr, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Adjust IP header*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr));

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=5;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=1;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/* Adjust Sequence and Acknowledgment numbers and window*/
		tcph->seq=htonl(++state.seq_num);
		tcph->ack_seq=htonl(state.ack_num);
		tcph->window=htons(WIN_FACTOR);

		/*Update Sequence Number to include the fin packet in the sequence number space*/
		state.seq_num++;

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,tcph->doff*4,
								(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/* Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);

	/* 2)Create Fin,Ack Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.ts.tv_usec+=1000; /*After the previous packet*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(ptr, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Update IP header, including swapping source and destination addresses*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		temp=iph->saddr;
		iph->saddr=iph->daddr;
		iph->daddr=temp;
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr));

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=5;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=1;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*Adjust Sequence and Acknowledgment numbers and window*/
		tcph->seq=htonl(state.ack_num++);
		tcph->ack_seq=htonl(state.seq_num);
		tcph->window=htons(WIN_FACTOR);

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,sizeof(struct tcphdr),
								(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/*Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);

	/* 3)Create Ack Packet*/
		/*make sure the packet is all zero*/
		memset(data, 0, MAX_PACKET);
		ptr=data;

		/*Set the libpcap header*/
		nh.caplen=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.len=sizeof(struct ether_header) +sizeof(struct iphdr)+sizeof(struct tcphdr);
		nh.ts.tv_usec+=1000; /*After the previous packet*/

		/* Copy Ethernet and IP headers from private data area*/
		/* These are headers from the first packet in the capture*/
		memcpy(ptr, eip->od, sizeof(struct ether_header)+ sizeof(struct iphdr));

		/*Update IP header*/
		iph= (struct iphdr *) (ptr + sizeof(struct ether_header));
		iph->protocol=6;
		iph->check=htonl(0);
		iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct tcphdr));

		/*Compute IPv4 Checksum*/
		iph->check=0;
		iph->check=ipv4_chksum((u_char*)iph,iph->ihl*4);

		/*Build TCP header*/
		ptr+=sizeof(struct ether_header)+ sizeof(struct iphdr);
		tcph=(struct tcphdr*)ptr;
		tcph->source=htons(1113);
		tcph->dest=htons(1113);
		tcph->doff=5;
		tcph->check=htonl(0);
		tcph->urg_ptr=0;
		tcph->urg=0;
		tcph->psh=0;
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*Adjust Sequence and Acknowledgment numbers and window*/
		tcph->seq=htonl(state.seq_num++);
		tcph->ack_seq=htonl(state.ack_num);
		tcph->window=htons(WIN_FACTOR);

		/*Compute TCP Checksum*/
		tcph->check=0;
		tcph->check=ipv4_pseudohdr_chksum((u_char*)tcph,tcph->doff*4,
								(u_char*)&iph->daddr, (u_char*)&iph->saddr, iph->protocol);

		/*Save To Packet Capture*/
		pcap_dump((u_char*)state.out,&nh, data);
	return 0;
}
