/******************************************************************************
Utility to convert a LTP flow to a TCP flow for LTP analysis via tcptrace.

Copyright (C) 2013  Samuel Jero <sj323707@ohio.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Author: Samuel Jero <sj323707@ohio.edu>
Date: 06/2013

Notes:
	1)Only handles one LTP "connection". There isn't a good way to separate
		different LTP "connections" from new sessions of the same "connection".
		Use Tcpdump filters to separate connections. Libpcap filtering could also
		be added in ltp2tcp.
	2)Uses some special types from Linux (u_char, u_int32_t)
******************************************************************************/
#include "ltp2tcp.h"

#define LTPTRACE_VERSION 0.3
#define COPYRIGHT_YEAR 2013

int debug=0;
int MAX_SESSION=500000;



/* Forward declarations*/
int init_state();
int free_state();
void handle_packet(u_char *user, const struct pcap_pkthdr *h, const u_char *bytes);
int convert_packet(struct pcap_pkthdr *h, const u_char *odata, u_char *ndata, int datalength, int newlength);
int handle_data(const u_char* odata, struct ltp_hdr_d* ltph, struct tcphdr *tcph, int* dlen, int rbuf);
int handle_report(const u_char* odata, struct ltp_hdr_d* ltph, struct tcphdr *tcph, u_char* tcpopt, int* dlen, int rbuf);
int claim2sack(const u_char* odata, int ses_id, struct ltp_hdr_d* ltph, struct tcphdr *tcph, u_char* tcpopt, int* dlen, int rbuf);
int seq_from_session(struct ltp_hdr_d* ltph, struct tcphdr *tcph);
int ack_from_session(struct ltp_hdr_d* ltph, struct tcphdr *tcph);
void version();
void usage();


/*Parse commandline options and open files*/
int main(int argc, char *argv[])
{
	char ebuf[200];
	char *erbuffer=ebuf;
	char *lfile=NULL;
	char *tfile=NULL;
	char *type=NULL;
	char *temp;
	int  tmp;

	/*parse commandline options*/
	if(argc < 2){
		usage();
	}

	/*loop through commandline options*/
	for(int i=1; i < argc; i++){
		if(argv[i][0]!='-' || (argv[i][0]=='-' && strlen(argv[i])==1)){
			if(lfile==NULL  || argv[i][0]=='-'){
				/*assign first non-dash (or only dash) argument to the dccp file*/
				lfile=argv[i];
			}else{
				if(tfile==NULL){
					tfile=argv[i]; /*assign second non-dash argument to the dccp file*/
				}else{
					usage();
				}
			}
		}else{
			if(argv[i][1]=='v' && strlen(argv[i])==2){ /*debug option*/
				debug++;
			}
			if(argv[i][1]=='V' && strlen(argv[i])==2){ /*Version option*/
				version();
			}
			if(argv[i][1]=='h'&& strlen(argv[i])==2){ /* help*/
				usage();
			}
			if(argv[i][1]=='t'){ /*Encapsulation option*/
				type=&argv[i][2];
			}
			if(argv[i][1]=='b'){ /* Block size option*/
				tmp=atoi(&argv[i][2]);
				if(tmp>0){
					MAX_SESSION=tmp;
				}else{
					usage();
				}
			}
			if(argv[i][1]=='s'){ /*Session range option*/
					state.ses_min=strtol(&argv[i][2],&temp,10);
					temp++;
					state.ses_max=strtol(temp,NULL,10);
			}
		}
	}
	
	if(lfile==NULL || tfile==NULL || type==NULL){
		usage();
	}

	if(state.ses_min<=0 || state.ses_max<=0){
		state.ses_min=-1;
		state.ses_max=-1;
	}

	/*all options validated*/

	if(debug){
		dbgprintf(1,"Input file: %s\n", lfile);
		dbgprintf(1,"Output file: %s\n", tfile);
		dbgprintf(1,"Encapsulation: %s\n", type);
		dbgprintf(1,"Block Size: %i\n", MAX_SESSION);
		if(state.ses_min>=0){
			dbgprintf(1, "Session Range: %i-%i\n", state.ses_min, state.ses_max);
		}else{
			dbgprintf(1, "Session Range: all\n");
		}
	}

	/*determine encapsulation type*/
	encap_sel(type);

	/*attempt to open input file*/
	state.in=pcap_open_offline(lfile, erbuffer);
	if(state.in==NULL){
		printf("Error opening input file\n");
		exit(1);
	}

	/*attempt to open output file*/
	state.out=pcap_dump_open(state.in,tfile);
	if(state.out==NULL){
		printf("Error opening output file\n");
		exit(1);
	}

	/*setup state*/
	if(init_state()){
		exit(1);
	}

	/*process packets*/
	u_char *user=(u_char*)state.out;
	pcap_loop(state.in, -1, handle_packet, user);

	/*add correct TCP termination, if needed*/
	if(state.en_ops && state.en_ops->fin){
			(*state.en_ops->fin)();
			free(state.en_priv);
	}
	
	/*close files*/
	pcap_close(state.in);
	pcap_dump_close(state.out);
return 0;
}

/*initialize the converter state*/
int init_state()
{
	state.seq_num=0;
	state.ack_num=0;
	state.win=0;
	state.started=0;
	state.seq_ses_id=0;
	state.ses.size=TBL_SZ;
	state.ses.max=0;
	state.en_priv=NULL;
	state.sender_id=-1;
	state.ses.table=malloc(TBL_SZ*sizeof(struct tbl));
	if(state.ses.table==NULL){
		dbgprintf(0,"Error: Couldn't allocate Memory!\n");
		exit(1);
	}
return 0;
}

/*cleanup state*/
int free_state()
{
	struct cl* tmp;
	struct cl* f;

	/*clean up the session table*/
	if(state.ses.table){

		/*cleanup the claims in each session*/
		for(int i=0; i < MAX_SESSION; i ++){
			if(state.ses.table[i].a_lst!=NULL){
				tmp=state.ses.table[i].a_lst;
				while(tmp){
					f=tmp;
					tmp=tmp->next;
					free(f);
				}
			}
		}

		free(state.ses.table);
	}
	return 0;
}

/*call back function for pcap_loop--do basic packet handling*/
void handle_packet(u_char *user, const struct pcap_pkthdr *h, const u_char *bytes)
{
	u_char 				*ndata;
	u_char 				*nptr;
	const u_char 		*optr=bytes;
	int 				length;
	int					nlength;
	int 				tlen;
	struct pcap_pkthdr	nh;

	/*create new libpcap header*/
	memcpy(&nh, h, sizeof(struct pcap_pkthdr));

	/*create buffer for new packet*/
	nptr=ndata=malloc(MAX_PACKET);
	if(ndata==NULL){
		dbgprintf(0,"Error: Couldn't allocate Memory\n");
		exit(1);
	}

	/*make sure the packet is all zero*/
	memset(ndata, 0, MAX_PACKET);
	length=h->caplen;
	nlength=MAX_PACKET;

	/*handle encapsulations*/
	if(state.en_ops && state.en_ops->pre && state.en_ops->post){
		if((*state.en_ops->pre)(&nh, &optr, &nptr, &length, &nlength)<0){
			return;
		}
		tlen=convert_packet(&nh, optr, nptr, length, nlength);
		if(tlen <0){
			free(ndata);
			return;
		}
		if((*state.en_ops->post)(tlen, ndata)<0){
			return;
		}
	}else{
		return;
	}

	/*save packet*/
	pcap_dump(user,&nh, ndata);

	free(ndata);
return;
}

/*do all the ltp to tcp conversions---returns length of TCP segment*/
int convert_packet(struct pcap_pkthdr *h, const u_char *odata, u_char *ndata, int datalength, int newlength)
{	
	u_char* 			ncur=ndata;
	struct tcphdr 		*tcph;
	u_char* 			tcpopt;
	u_char				ctl;
	struct ltp_hdr_d 	ltph;
	int 				eaten;
	int					tmp;
	int					hext;
	int					dlen=0;
	int					helen;

	/*safety checks*/
	if(h==NULL || odata==NULL || ndata==NULL|| datalength<1|| newlength < sizeof(struct tcphdr)){
			return -1;
	}

	/*cast TCP header pointers*/
	tcph=(struct tcphdr*)ncur;
	tcpopt=ncur+ sizeof(struct tcphdr);

	/*set TCP standard features*/
	tcph->source=htons(1113);
	tcph->dest=htons(1113);
	tcph->doff=5;
	tcph->check=htonl(0);
	tcph->urg_ptr=0;
	tcph->urg=0;
	tcph->psh=0;

	/*Grab LTP control/version byte*/
	ltph.vers_ctl=ctl=*odata;
	odata++;
	datalength--;

	/* check LTP version*/
	if((ctl & 0xF0) !=0){
		dbgprintf(1,"Error: Invalid Packet Version\n");
		return -1;
	}

	/* Parse LTP Session ID*/
	ltph.sender_id=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	datalength-=eaten;
	if(ltph.sender_id<0 || datalength<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	ltph.session_num=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	datalength-=eaten;
	if(ltph.session_num<0 || datalength<0){
			dbgprintf(1,"Error: Bad SDNV!\n");
			return -1;
	}

	/*Filter sessions*/
	if(state.ses_min>=0){
		if((ltph.session_num < state.ses_min)|| (ltph.session_num > state.ses_max)){
			dbgprintf(3, "Filtering session: %i\n", ltph.session_num);
			return -1;
		}
	}

	/*Check LTP Sender*/
	/* We don't want to be confused if we have multiple LTP connections in this capture*/
	if(state.sender_id==-1){
		state.sender_id=ltph.sender_id;
	}else{
		if(state.sender_id!=ltph.sender_id){
			dbgprintf(1,"Note: Packet from different Sender!!\n");
			return -1;
		}
	}

	/*Parse and Remove LTP header extensions*/
	helen=0;
	ltph.extensions=*odata;
	odata++;
	hext=ltph.extensions >> 4;
	while(hext>0){
		tmp=evaluate_sdnv(odata+1, &eaten);
		if(tmp==0 || tmp < eaten || tmp < 0 || datalength < tmp){
			dbgprintf(1,"Error: Packet has bad extension length!\n");
			return -1;
		}
		/* ignore all extensions*/
		odata+=tmp;
		helen+=tmp;
		datalength-=tmp;
		hext--;
	}



	/*make changes by packet type*/
	if((ctl & 0x0C)==0x0){//RED DATA
		dbgprintf(2,"Packet Type: RED DATA\n");

		/*Initialize the TCP connection, if this is the start*/
		if(state.started==0){
			state.started=1;
			if(state.en_ops && state.en_ops->handshake){
				(*state.en_ops->handshake)(h);
			}
		}
		
		if(handle_data(odata, &ltph, tcph, &dlen, datalength)<0){
			return -1;
		}

		/*Finalize TCP Header*/
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*check length*/
		if(newlength < dlen + tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Add data*/
		memset(ndata + tcph->doff*4, 'A', dlen);

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=tcph->doff*4+dlen;
		h->caplen+=tcph->doff*4+dlen;
		return tcph->doff*4+dlen;
	}

	if((ctl & 0x0C)==0x4){//GREEN DATA
			dbgprintf(2,"Packet Type: GREEN DATA\n");

		/*Initialize the TCP connection, if this is the start*/
		if(state.started==0){
			state.started=1;
			if(state.en_ops && state.en_ops->handshake){
					(*state.en_ops->handshake)(h);
			}
		}

		if(handle_data(odata, &ltph, tcph, &dlen, datalength)<0){
			return -1;
		}

		/*Finalize TCP Header*/
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*check length*/
		if(newlength < dlen + tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Add data*/
		memset(ndata+ tcph->doff*4, 'A', dlen);

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=tcph->doff*4+dlen;
		h->caplen+=tcph->doff*4+dlen;
		return tcph->doff*4+dlen;
	}

	if((ctl & 0x0F)==0x8){//REPORT SEGMENT
		dbgprintf(2,"Packet Type: REPORT SEGMENT\n");

		/*check length*/
		if(newlength < 4*8+4+sizeof(struct tcphdr)){
			dbgprintf(1, "Error: Indicated Packet Length is may be beyond end of buffer!\n");
			return -1;
		}

		if(handle_report(odata, &ltph, tcph, tcpopt, &dlen, datalength)<0){
			return -1;
		}

		/*Finalize TCP Header*/
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*check length*/
		if(newlength < tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=tcph->doff*4+dlen;
		h->caplen+=tcph->doff*4+dlen;
		return tcph->doff*4+dlen;
	}

	if((ctl & 0x0F)==0x9){//REPORT ACK
		dbgprintf(2,"Packet Type: REPORT ACK\n");

		/*Get Seq and Ack Numbers*/
		if(seq_from_session(&ltph, tcph)<0){
			return -1;
		}
		tcph->ack_seq=htonl(state.ack_num);

		/*Finalize TCP Header*/
		tcph->window=htons(WIN_FACTOR);
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=0;
		tcph->ack=1;

		/*check length*/
		if(newlength < tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=sizeof(struct tcphdr);
		h->caplen+=sizeof(struct tcphdr);
		return sizeof(struct tcphdr);
	}

	if((ctl & 0x0D)==0xC){ //CANCEL SEGMENT
		dbgprintf(2,"Packet Type: CANCEL SEGMENT\n");

		/*Get Seq and Ack Numbers*/
		if((ctl & 0x02)==0x00){
			//Sender cancel
			if(seq_from_session(&ltph, tcph)<0){
				return -1;
			}
			tcph->ack_seq=htonl(state.ack_num);
		}else{
			//Receiver cancel
			if(ack_from_session(&ltph, tcph)<0){
				return -1;
			}
			tcph->seq=htonl(state.ack_num++);
		}

		/*Finalize TCP Header*/
		tcph->window=htons(0);
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=1;
		tcph->ack=0;

		/*check length*/
		if(newlength < tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=sizeof(struct tcphdr);
		h->caplen+=sizeof(struct tcphdr);
		return sizeof(struct tcphdr);
	}

	if((ctl & 0x0D)==0xD){//CANCEL ACK
		dbgprintf(2,"Packet Type: CANCEL ACK\n");

		/*Get Seq and Ack Numbers*/
		if((ctl & 0x02)==0x00){
			//Sender ACK
			if(seq_from_session(&ltph, tcph)<0){
				return -1;
			}
			tcph->ack_seq=htonl(state.ack_num);
		}else{
			//Receiver ACK
			if(ack_from_session(&ltph, tcph)<0){
				return -1;
			}
			tcph->seq=htonl(state.ack_num++);
		}

		/*Finalize TCP Header*/
		tcph->window=htons(0);
		tcph->fin=0;
		tcph->syn=0;
		tcph->rst=1;
		tcph->ack=0;

		/*check length*/
		if(newlength < tcph->doff*4){
			dbgprintf(1, "Error: Indicated Packet Length is beyond end of buffer!\n");
			return -1;
		}

		/*Adjust Lengths (packet and TCP segment)*/
		h->len+=sizeof(struct tcphdr);
		h->caplen+=sizeof(struct tcphdr);
		return sizeof(struct tcphdr);
	}

	/*Anything else is bad!*/
	dbgprintf(1,"Error: Invalid Packet Type\n");
return -1;
}

/* Determine Sequence and Acknowledgment Numbers for Data Segments*/
int handle_data(const u_char* odata, struct ltp_hdr_d* ltph, struct tcphdr *tcph, int* dlen, int rbuf)
{
	int eaten;
	int i;
	int found;

	/* Get LTP Client Service ID*/
	ltph->cls_id=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->cls_id <0 || rbuf <0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get LTP Data Offset*/
	ltph->offset=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->offset <0 || rbuf <0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get LTP Data Length*/
	ltph->length=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->length <0 || rbuf <0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}
	
	/* find LTP session */
	found=0;
	for(i=0; i < state.ses.size; i++){
		if(ltph->session_num==state.ses.table[i].num){
			found=1;
			break;
		}
	}
	
	/* Not Found, add it*/
	if(found==0){
		i=(state.seq_ses_id++)%state.ses.size;
		state.ses.table[i].num=ltph->session_num;
		state.ses.table[i].seq_id=state.seq_ses_id;
		if(state.seq_ses_id==1){
			/*First session*/
			state.ses.table[i].start=state.seq_num;
		}else{
			state.ses.table[i].start=state.seq_num + MAX_SESSION - state.ses.table[(state.seq_ses_id-2)%state.ses.size].length;
		}
	}

	/* Compute TCP Sequence Number */
	tcph->seq=htonl(state.ses.table[i].start + ltph->offset);
	if(state.ses.table[i].start + ltph->offset+ltph->length > state.seq_num){
		state.seq_num=state.ses.table[i].start + ltph->offset + ltph->length;
	}

	/* Update LTP Session Length*/
	if(ltph->offset+ltph->length > state.ses.table[i].length){
		state.ses.table[i].length=ltph->offset+ltph->length;
	}

	/* Computer TCP Ack Number*/
	tcph->ack_seq=htonl(state.ack_num);

	/* Get Amount of Data in LTP Segment*/
	*dlen=ltph->length;

	/* Compute TCP Window*/
	tcph->window=htons(WIN_FACTOR);
return 0;
}

/* Get Sequence and Acknowledgment Number for LTP Reports. Add TCP SACKS*/
int handle_report(const u_char* odata, struct ltp_hdr_d* ltph, struct tcphdr *tcph, u_char* tcpopt, int* dlen, int rbuf)
{
	int eaten;
	int found;
	int i;
	int data_recv;

	/*Set TCP Data Length*/
	*dlen=0;

	/*Parse LTP header*/
	/*Get LTP Report ID*/
	ltph->report=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->report <0 || rbuf<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get LTP Checkpoint ID*/
	ltph->checkpoint=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->checkpoint <0 || rbuf<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get LTP Report Upper Bound*/
	ltph->ubound=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->ubound <0 || rbuf<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get LTP Report Lower Bound*/
	ltph->lbound=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->lbound <0 || rbuf<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/*Get Number of LTP Reception Claims*/
	ltph->rclaims=evaluate_sdnv(odata, &eaten);
	odata+=eaten;
	rbuf-=eaten;
	if(ltph->rclaims <0 || rbuf<0){
		dbgprintf(1,"Error: Bad SDNV!\n");
		return -1;
	}

	/* find LTP session */
	found=0;
	for(i=0; i < state.ses.size; i++){
		if(ltph->session_num==state.ses.table[i].num){
			found=1;
			break;
		}
	}

	/* LTP session Not Found*/
	if(found==0){
		if(debug){
			dbgprintf(1,"Error: Session for Report Not Found!\n");
		}
		return -1;
	}

	/* Add TCP SACKS*/
	data_recv=claim2sack(odata, i, ltph, tcph, tcpopt, dlen, rbuf);
	if(data_recv<0){
		return -1;
	}

	/*Add TCP sequence number*/
	tcph->seq=htonl(state.ack_num);
	
	/* Compute TCP Window*/
	tcph->window=htons(WIN_FACTOR);

	/* Add TCP Ack Number*/
	tcph->ack_seq=htonl(state.ses.table[i].start + data_recv+1);
return 0;
}

/* Convert LTP Reception Claims into TCP SACKS*/
int claim2sack(const u_char* odata, int ses_id,struct ltp_hdr_d* ltph, struct tcphdr *tcph, u_char* tcpopt, int* dlen, int rbuf)
{
	int 		coffset;
	int 		clength;
	int 		eaten;
	int 		max;
	int			done;
	u_char 		*olen;
	u_int32_t* 	t_block;
	struct cl* 	tmp=NULL;

	/*Add the Start of the TCP SACK Option*/
	*tcpopt=5;
	tcpopt++;
	olen=tcpopt;
	tcpopt++;
	*olen=2;
	*dlen=0;
	tcph->doff+=1;

	/* Cast to TCP SACK BLOCK pointer*/
	t_block=(u_int32_t*)tcpopt;
	
	/* Loop over at most 4 Claims*/
	if(ltph->rclaims>4){
		max=4;
	}else{
		max=ltph->rclaims;
	}

	for(int i=0; i < max; i++){

		/* Read LTP Claim*/
		coffset=evaluate_sdnv(odata, &eaten);
		odata+=eaten;
		rbuf-=eaten;
		if(coffset <0 || rbuf<0){
			dbgprintf(1,"Error: Bad SDNV!\n");
			return -1;
		}
		clength=evaluate_sdnv(odata, &eaten);
		odata+=eaten;
		rbuf-=eaten;
		if(clength <0 || rbuf<0){
			dbgprintf(1,"Error: Bad SDNV!\n");
			return -1;
		}

		/* Add claim to session info*/
		if(!state.ses.table[ses_id].a_lst){
			state.ses.table[ses_id].a_lst=malloc(sizeof(struct cl));
			if(!state.ses.table[ses_id].a_lst){
				dbgprintf(0,"Error: Couldn't allocate Memory\n");
				exit(1);
			}
			tmp=state.ses.table[ses_id].a_lst;
		}else{
			tmp=state.ses.table[ses_id].a_lst;
			while(tmp->next){
				tmp=tmp->next;
			}
			tmp->next=malloc(sizeof(struct cl));
			if(!tmp->next){
				dbgprintf(0,"Error: Couldn't allocate Memory\n");
				exit(1);
			}
		}
		tmp->offset=coffset;
		tmp->length=clength;

		/*Create TCP SACK*/
		*t_block=htonl(state.ses.table[ses_id].start+coffset); //Left Side
		t_block++;
		*t_block=htonl(state.ses.table[ses_id].start+coffset+clength+1); //Right Side
		t_block++;

		/*Update TCP header length*/
		*olen+=8;
		tcph->doff+=2;
	}

	/*Calculate Ack number*/
	done=0;
	coffset=0;
	while(!done){
		tmp=state.ses.table[ses_id].a_lst;
		done=1;
		while(tmp){
			if((tmp->offset<=coffset) && ((tmp->offset+tmp->length) > coffset)){
				coffset=tmp->offset+tmp->length;
				done=0;
			}
			tmp=tmp->next;
		}
	}
	state.ses.table[ses_id].acked=coffset;
return coffset;
}

/* Get a TCP seq number from the LTP Session ID for Report ACK and Session Cancellation*/
int seq_from_session(struct ltp_hdr_d* ltph, struct tcphdr *tcph)
{
	int i;
	int found;

	/*Find LTP session */
	found=0;
	for(i=0; i < state.ses.size; i++){
		if(ltph->session_num==state.ses.table[i].num){
			found=1;
			break;
		}
	}

	/*LTP session Not Found*/
	if(found==0){
		dbgprintf(1,"Error: Session for Segment Not Found!\n");
		return -1;
	}

	/* Add TCP Sequence Number */
	tcph->seq=htonl(state.ses.table[i].start + state.ses.table[i].length);
	/* No length, so don't update state.ses.table[i].length*/
return 0;
}

/* Get an TCP Ack number from the LTP session ID for session cancellation*/
int ack_from_session(struct ltp_hdr_d* ltph, struct tcphdr *tcph)
{

	int i;
	int found;

	/* find LTP session */
	found=0;
	for(i=0; i < state.ses.size; i++){
		if(ltph->session_num==state.ses.table[i].num){
			found=1;
			break;
		}
	}

	/* LTP session Not Found*/
	if(found==0){
		if(debug){
			dbgprintf(1,"Error: Session for Segment Not Found!\n");
		}
		return -1;
	}
	
	/* Add TCP Ack Number*/
	tcph->ack_seq=htonl(state.ses.table[i].start + state.ses.table[i].acked+1);
return 0;
}

void version()
{
	dbgprintf(0, "ltp2tcp version %.1f\n",LTPTRACE_VERSION);
	dbgprintf(0, "Copyright (C) %i Samuel Jero <sj323707@ohio.edu>\n",COPYRIGHT_YEAR);
	dbgprintf(0, "This program comes with ABSOLUTELY NO WARRANTY.\n");
	dbgprintf(0, "This is free software, and you are welcome to\n");
	dbgprintf(0, "redistribute it under certain conditions.\n");
	exit(0);
}

/*Usage information for program*/
void usage()
{
	dbgprintf(0,"Usage: ltp2tcp -t{encapsulation} [-v] [-V] [-h] [-b{block_size}] [-s{start session}-{end session}] ltp_file tcp_file\n");
	dbgprintf(0, "          -v   verbose. May be repeated for additional verbosity.\n");
	dbgprintf(0, "          -V   Version information\n");
	dbgprintf(0, "          -h   Help\n");
	dbgprintf(0, "          -t   type of encapsulation (udp,dccp,sll)\n");
	dbgprintf(0, "          -b   LTP block size (NOT the bundle size) that is being used over this connection\n");
	dbgprintf(0, "          -s   Requests a graph of only the specified range of sessions\n");
	exit(0);
}

/*Debug Printf*/
void dbgprintf(int level, const char *fmt, ...)
{
    va_list args;
    if(debug>=level){
    	va_start(args, fmt);
    	vfprintf(stderr, fmt, args);
    	va_end(args);
    }
}
